package com.sonydna.common.extensions

import java.io.File

import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.BitmapFactory.Options
import android.graphics.Canvas
import android.graphics.Point
import android.graphics.Rect
import android.graphics.RectF
import android.view.Display
import android.widget.ImageView

import com.sonydna.common.annotation.functional_programming.AsPure
import com.sonydna.common.annotation.functional_programming.Immutable
import com.sonydna.common.annotation.null_analysis.NonNullByDefault

@Immutable
@NonNullByDefault
final class ScDimension {
    public final int w
    public final int h

    new(int w, int h) {
        this.w = w
        this.h = h
    }

    //	conversions ///////////////////////////////////////
    
    def static final getDimension(Rect      e) { new ScDimension(e.width, e.height) }
    def static final getDimension(RectF     e) { new ScDimension(e.width as int, e.height as int) }
    def static final getDimension(Options   e) { new ScDimension(e.outWidth, e.outHeight) }
    def static final getDimension(Bitmap    e) { new ScDimension(e.width, e.height) }
    def static final getDimension(Canvas    e) { new ScDimension(e.width, e.height) }
    def static final getDimension(ImageView e) { new ScDimension(e.width, e.height) }
    
    def static final ScDimension getDimensionByPhysical(File e) {
        val options = new BitmapFactory.Options => [ inJustDecodeBounds = true ]
        BitmapFactory.decodeFile(e.absolutePath, options)
        new ScDimension(options.outWidth, options.outHeight)
    }

    def static final ScDimension getDimensionByLogical(File e) {
    	throw new UnsupportedOperationException
    }

    @Deprecated
    def static final ScDimension getDimension(Display e) {
        return new ScDimension(e.width, e.height)
    }

    //	xtend operator overload ///////////////////////////
    
    @AsPure def final operator_equals         (ScDimension e) { w == e.w && h == e.h }
    @AsPure def final operator_notEquals      (ScDimension e) { w != e.w || h != e.h }
    @AsPure def final operator_lessThan       (ScDimension e) { w <  e.w && h <  e.h }
    @AsPure def final operator_lessEqualsThan (ScDimension e) { w <= e.w && h <= e.h }
    @AsPure def final operator_plus           (ScDimension e) { new ScDimension(w + e.w, h + e.h) }
    @AsPure def final operator_minus          (ScDimension e) { new ScDimension(w - e.w, h - e.h) }
    @AsPure def final operator_multiply       (float e)       { new ScDimension((w * e) as int, (h * e) as int) }
    @AsPure def final operator_divide         (float e)       { new ScDimension((w / e) as int, (h / e) as int) }

    ///////////////////////////////////////////////////////
    
    @AsPure override final ScDimension clone() {
        new ScDimension(w, h)
    }

    @AsPure def final int getLonger() {
        Math.max(w, h)
    }

    @AsPure def final int getShorter() {
        Math.min(w, h)
    }

    @AsPure def final int getSize() {
        w * h
    }

    @AsPure def final Point getCenter() {
        new Point(w / 2, h / 2)
    }

    @AsPure def final ScDimension swap() {
        new ScDimension(h, w)
    }

    @AsPure def final float fitInsideScale(ScDimension outerSize) {
        Math.min(outerSize.w as float / w, outerSize.h as float / h)
    }

    @AsPure def final float fitOutsideScale(ScDimension outerSize) {
        Math.max(outerSize.w as float / w, outerSize.h as float / h)
    }

    @AsPure def final ScDimension fitInside(ScDimension outerSize) {
        val scale = fitInsideScale(outerSize)
        new ScDimension((w * scale) as int, (h * scale) as int)
    }

    @AsPure def final ScDimension fitOutside(ScDimension outerSize) {
        val scale = fitOutsideScale(outerSize)
        new ScDimension((w * scale) as int, (h * scale) as int)
    }

    @AsPure def final int insideSampleSize(ScDimension requestSize) {
        var result = 1
        var w = this.w
        var h = this.h
        while (requestSize.w < w || requestSize.h < h) {
            w = w / 2
            h = h / 2
            result = result * 2
        }
        return result
    }

    @AsPure def final int outsideSampleSize(ScDimension requestSize) {
        var result = 1
        var w = this.w
        var h = this.h
        while (requestSize.w <= w && requestSize.h <= h) {
            w = w / 2
            h = h / 2
            result = result * 2
        }
        return Math.max(1, result / 2)
    }

    @AsPure def final int insideSampleSize(float requestScale) {
        return insideSampleSize(new ScDimension((w * requestScale) as int, (h * requestScale) as int))
    }

    @AsPure def final int outsideSampleSize(float requestScale) {
        return outsideSampleSize(new ScDimension((w * requestScale) as int, (h * requestScale) as int))
    }

	//	TODO refactoring //////////////////////////////////

    def final float getAspectRatio() {
        if (w == 0 || h == 0) return 9999
        if (w > h) w / h as float else h / w as float
    }

    def private static final float fitInsideScale(float w, float h, float outerW, float outerH) {
        return Math.min(outerW / w, outerH / h)
    }

    def private static final float fitOutsideScale(float w, float h, float outerW, float outerH) {
        return Math.max(outerW / w, outerH / h)
    }

    def static final ScDimension spaceSizeForMatchAspectRatio(int w, int h, int aspectRatioX, int aspectRatioY) {
        val scale = fitOutsideScale(aspectRatioX, aspectRatioY, w, h)
        val dx = ((aspectRatioX * scale) - w) as int / 2
        val dy = ((aspectRatioY * scale) - h) as int / 2
        return new ScDimension(dx, dy)
    }

    def static final float insideScale(float w, float h, float outerW, float outerH) {
        val fitInsideScale = fitInsideScale(w, h, outerW, outerH)
        return Math.min(fitInsideScale, 1)
    }
}
